// Worley noise, version 1, 5/8/2021
// By Jacob, cheeseisatrafficsignal.com
// Licensed under CC0, except libattopng by Michael Schwarz, which is licensed under MIT

// Usage: Up and Down arrow keys to move the view slice
//		  Space to play and pause the automatic movement
//        Escape to exit

// Dependences: SDL2 and libattopng

#include <SDL2/SDL.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "libattopng.h"
#include "display.h"

#define MAX_POINTS 90 // Has to be divisible by 3

// Animation variables
libattopng_t* frame = NULL;
int frame_counter = 0;
char *output_path = NULL;
bool save_animation = false;

// General variables
uint32_t render_color = 0xFFFFFFFF; // ARGB
uint32_t background_color = 0xFF3D3D3D;
int max_z = 900; // This has to be divisible by 3
int z_slice;
int min_z = 0;

bool is_running;
bool debug = false;
bool paused = false;
int previous_frame_time = 0;

typedef struct {
	int x;
	int y;
	int z;
	float weight;
} point_t;

point_t points[MAX_POINTS];

// Function to return a random integer between low (l) and high (h)
int random(int l, int h) {
	return (rand() % (h - l + 1))+l;
}

// Function to initialize the points for the Worley noise
void init_points() {
	for (int i=0;i<MAX_POINTS/3;i++) {
		point_t new_point = {		
			.x = random(50,window_width-50),
			.y = random(50,window_height-50),
			.z = random(min_z,max_z/3),
			.weight = (float)random(70,100)/100
		};
		point_t new_point_loop = new_point;
		new_point_loop.z+=max_z/3; // These additional points are here to make the animation loop
		points[i] = new_point;
		points[i+MAX_POINTS/3] = new_point_loop;
		new_point_loop.z+=max_z/3;
		points[i+(MAX_POINTS/3*2)] = new_point_loop;
	}	
}

void setup(void) {
	// Set up the color buffer
	color_buffer = (uint32_t*) malloc(sizeof(uint32_t) * window_width * window_height);
	color_buffer_texture = SDL_CreateTexture(
		renderer,
 		SDL_PIXELFORMAT_RGBA32,
 		SDL_TEXTUREACCESS_STREAMING,
		window_width,
		window_height
	);
	// Seed the random generator with the current time. I know, not great
	srand(time(0));
	// Initialize the points.
	init_points();
	// Starting z slice
	z_slice = max_z/3;
	// Frame to save animation. The same frame is updated and then exported each update
	frame = libattopng_new(window_width, window_height, PNG_RGBA);
	output_path = (char*)malloc(40 * sizeof(char));	
}

// SDL event and key press handling
void process_input(void) {
	SDL_Event event;
	while (SDL_PollEvent(&event)) {
		switch (event.type) {
			case SDL_QUIT:
				is_running = false;
				break;
			case SDL_KEYDOWN:
				if (event.key.keysym.sym == SDLK_ESCAPE) {
					is_running = false;
				}
				if (event.key.keysym.sym == SDLK_SPACE) {
					paused = !paused;
				}
				if (event.key.keysym.sym == SDLK_UP) {
					z_slice=(z_slice+10)%max_z;
				}
				if (event.key.keysym.sym == SDLK_DOWN) {
					z_slice=(z_slice-10)%max_z;
				}
				break;
		}
	}
}

void update(void) {
	// Wait until the target FPS milliseconds have elapsed
	int time_to_wait = FRAME_TARGET_TIME - (SDL_GetTicks() - previous_frame_time    );
	if (time_to_wait > 0 && time_to_wait <= FRAME_TARGET_TIME) {
		SDL_Delay(time_to_wait);
	}
	// Get delta time
	float delta_time = (SDL_GetTicks() - previous_frame_time) / 1000.0;
	previous_frame_time=SDL_GetTicks();
	// Make the worley noise
	for (int x=0;x<window_width; x++) {
		for (int y=0;y<window_height; y++) {
			float dist = 800;
			// Find the nearest point
			for (int p=0;p<MAX_POINTS;p++) {
				// Calculate the distance from the current pixel to the current point
				int dist_x = points[p].x - x;
				int dist_y = points[p].y - y;
				int dist_z = points[p].z - z_slice;
				float c_dist = sqrt(dist_x*dist_x+dist_y*dist_y+dist_z*dist_z);
				c_dist*=points[p].weight;
				if (c_dist<dist) dist = c_dist;	// If the current distance is less than the previous distance, replace it
			}
			float percent = 1-(dist/800)/0.2;
			if (percent<0) percent=0;
			if (percent>1) percent=1;
			// Scale the color to be the original color close to the point and black further away
			uint32_t alpha = (render_color & 0xFF000000);
			uint32_t red = (render_color & 0x00FF0000)*percent;
			uint32_t green = (render_color & 0x0000FF00)*percent;
			uint32_t blue = (render_color & 0x000000FF)*percent;
			uint32_t color = alpha | (red & 0x00FF0000) | (green & 0x0000FF00) | (blue & 0x000000FF);
			// Finally, draw the pixel
			draw_pixel(x,y,color);
			if (save_animation && !paused) // Also draw the pixel to the PNG frame if save_animation is true
				libattopng_set_pixel(frame, x, y, color);
		}
	}
	if (!paused) { // Only move the z_slice and output the animation frames if not paused	
		z_slice += 40*delta_time;
		if (z_slice>=max_z/3*2) { // Stop the slice movement once it reaches the beginning. This way the output frames will be a perfect loop
			z_slice=max_z/3;
			paused = true;
		}
		if (save_animation) {
			snprintf(output_path, 25, "frames/frame_%d.png", frame_counter); // Yeah, this isn't great
			libattopng_save(frame, output_path);
			frame_counter++;
		}
	}
}

void render(void) {
	if (debug) {
		// Show a top down view of where the points are positioned
		for (int i=0;i<MAX_POINTS;i++) {
			draw_rect(points[i].z,points[i].y,5,5,0xFFFF00FF);
		}
	}
	render_color_buffer();
	clear_color_buffer(background_color);
	SDL_RenderPresent(renderer);
}

void free_resources(void) {
	free(color_buffer);
	libattopng_destroy(frame);
}

int main(void) {
	is_running = initialize_window();
	setup();
	while (is_running) {
		process_input();
		update();
		render();
	}
	destroy_window();
	free_resources();
	return 0;
}
